# -*- coding: utf-8 -*-
# Copyright 2023 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Common code for BumpDeps Unit Tests**
"""

from contextlib import contextmanager
import difflib
from dataclasses import dataclass
import unittest
from pathlib import Path
from tempfile import NamedTemporaryFile

import responses


EXAMPLE = """
[project]
dependencies = [
    'packaging == 1.0.0',
    'requests <= 2.0',
    'enlighten == 1.11.1',
    'prefixed',
]
[project.optional-dependencies]
opt1 = [
    'jinxed == 1.2.0',
    'prefixed ~= 0.3.2',  # Non-actionable comment
]
opt2 = [
    'sphinx ~= 5.3',
]
opt3 = [
    'blessed >= 1.0, < 1.16',
]
opt4 = [
    'pydantic == 1.2',  # bumpdeps: ignore
    'fastapi == 0.76.0',  # bumpdeps: ignore-until=2200-01-01
    'six == 1.7.1',  # bumpdeps: ignore-until=1984-01-01
]
opt5 = [
    'pylint == 2.15.10'
]

"""

# File diff for base dependencies
DIFF_BASE = (
    "-     'packaging == 1.0.0',",
    '+     "packaging == 23.1",',
    "-     'requests <= 2.0',",
    '+     "requests <= 2.28.2",',
)

# File diff for all extras
DIFF_EXTRAS = (
    "-     'prefixed ~= 0.3.2',  # Non-actionable comment",
    '+     "prefixed ~= 0.6.0",  # Non-actionable comment',
    "-     'sphinx ~= 5.3',",
    '+     "sphinx ~= 6.1",',
    "-     'blessed >= 1.0, < 1.16',",
    '+     "blessed < 1.19.1, >= 1.0",',
    "-     'six == 1.7.1',  # bumpdeps: ignore-until=1984-01-01",
    '+     "six == 1.16.0",  # bumpdeps: ignore-until=1984-01-01',
)


@ dataclass
class Result:
    """
    Simple container to return data from write_and_diff()
    """

    diff: tuple = tuple()
    file: str = None

    # Only for main()
    exit_code: int = 0
    logs: list = None
    stderr: str = ''
    stdout: str = ''


@contextmanager
def write_and_diff(text):
    """
    Write data to a temporary file and return a summary of the differences on close
    """

    test_file = None
    result = Result()

    try:
        test_file = NamedTemporaryFile(mode='w+', encoding='utf-8')
        test_file.write(text)
        test_file.flush()
        result.file = test_file.name
        yield result

    finally:

        new_text = Path(test_file.name).read_text(encoding='utf-8')
        diff = difflib.ndiff(text.splitlines(), new_text.splitlines())
        result.diff = tuple(line for line in diff if line[0] in '-+')
        test_file.close()


class MockedResponse(unittest.TestCase):
    """
    Test case class with mocked PyPI responses
    """

    def setUp(self) -> None:

        responses.get('https://pypi.org/pypi/requests/json',
                      json={'info': {'version': '2.28.2'}}, status=200)

        responses.get('https://pypi.org/pypi/packaging/json',
                      json={'info': {'version': '23.1'}}, status=200)

        responses.get('https://pypi.org/pypi/enlighten/json',
                      json={'info': {'version': '1.11.1'}}, status=200)

        responses.get('https://pypi.org/pypi/blessed/json',
                      json={'info': {'version': '1.19.1'}}, status=200)

        responses.get('https://pypi.org/pypi/prefixed/json',
                      json={'info': {'version': '0.6.0'}}, status=200)

        responses.get('https://pypi.org/pypi/sphinx/json',
                      json={'info': {'version': '6.1.2'}}, status=200)

        responses.get('https://pypi.org/pypi/jinxed/json',
                      json={'info': {'version': '1.2.0'}}, status=200)

        responses.get('https://pypi.org/pypi/six/json',
                      json={'info': {'version': '1.16.0'}}, status=200)

        responses.get('https://pypi.org/pypi/pylint/json',
                      json={'info': {'version': '2.15.10'}}, status=200)

        responses.get('https://pypi.org/pypi/no-such-package/json',
                      json={"message": "Not Found"}, status=404)

        responses.get('https://pypi.org/pypi/invalid-json/json',
                      body='Hello!', status=200)

        responses.get('https://pypi.org/pypi/unexpected-json/json',
                      json={'version': '1.2.3'}, status=200)



# -*- coding: utf-8 -*-
# Copyright 2023 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**BumpDeps CLI Unit Tests**
"""

from contextlib import contextmanager, redirect_stderr, redirect_stdout
from io import StringIO

import responses

import bumpdeps
from tests import DIFF_BASE, DIFF_EXTRAS, EXAMPLE, MockedResponse, write_and_diff


class MockedCLI(MockedResponse):
    """
    Provides additional methods to simplify writing CLI tests
    """

    @contextmanager
    def write_and_diff(self, text):
        """
        Write data to a temporary file
        Return a summary of the differences and system outputs on close
        """

        with write_and_diff(text) as result:
            with self.assertRaises(SystemExit) as sys_exit:
                with redirect_stderr(StringIO()) as err, redirect_stdout(StringIO()) as out:
                    with self.assertLogs('bumpdeps', level='ERROR') as logs:
                        yield result

        result.exit_code = sys_exit.exception.code
        result.logs = logs.output
        result.stderr = err.getvalue()
        result.stdout = out.getvalue()

    def check_result_error(self, result, regex):
        """
        Check result object for single logged exception
        """

        self.assertEqual(result.exit_code, 8)
        self.assertEqual(len(result.logs), 1)
        self.assertRegex(result.logs[0], regex)
        self.assertEqual(result.stderr, '')
        self.assertEqual(result.stdout, '')
        self.assertEqual(result.diff, ())

    def check_no_error(self, result):
        """
        Checks result object to make sure there are no errors
        """

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(len(result.logs), 0)
        self.assertEqual(result.stderr, '')


class TestMain(MockedCLI):
    """
    Tests for CLI entry point
    """

    @responses.activate
    def test_all(self):
        """Update all dependencies"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file, '--all'))

        self.check_no_error(result)
        self.assertEqual(
            result.stdout,
            (
                'Base Dependencies:\n'
                '    <-- packaging == 1.0.0\n'
                '    --> packaging == 23.1\n'
                '    <-- requests <= 2.0\n'
                '    --> requests <= 2.28.2\n'
                'Optional Dependencies:\n'
                '  opt1:\n'
                '    <-- prefixed ~= 0.3.2\n'
                '    --> prefixed ~= 0.6.0\n'
                '  opt2:\n'
                '    <-- sphinx ~= 5.3\n'
                '    --> sphinx ~= 6.1\n'
                '  opt3:\n'
                '    <-- blessed < 1.16, >= 1.0\n'
                '    --> blessed < 1.19.1, >= 1.0\n'
                '  opt4:\n'
                '    <-- six == 1.7.1\n'
                '    --> six == 1.16.0\n'
            )
        )
        self.assertEqual(result.diff, DIFF_BASE + DIFF_EXTRAS)

    @responses.activate
    def test_base_no_deps(self):
        """No dependencies listed"""

        with self.write_and_diff('[project]\n') as result:
            bumpdeps.main(('-f', result.file))

        self.check_no_error(result)
        self.assertEqual(result.stdout, 'No updates required\n')
        self.assertEqual(result.diff, ())

    @responses.activate
    def test_base_only(self):
        """Update only base dependencies"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file))

        self.check_no_error(result)
        self.assertEqual(
            result.stdout,
            (
                'Base Dependencies:\n'
                '    <-- packaging == 1.0.0\n'
                '    --> packaging == 23.1\n'
                '    <-- requests <= 2.0\n'
                '    --> requests <= 2.28.2\n'
                'Optional Dependencies:\n'
            )
        )
        self.assertEqual(result.diff, DIFF_BASE)

    @responses.activate
    def test_dry_run(self):
        """Updates returns, but file is not updated"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file, '--dry-run'))

        self.check_no_error(result)
        self.assertEqual(
            result.stdout,
            (
                'Base Dependencies:\n'
                '    <-- packaging == 1.0.0\n'
                '    --> packaging == 23.1\n'
                '    <-- requests <= 2.0\n'
                '    --> requests <= 2.28.2\n'
                'Optional Dependencies:\n'
            )
        )
        self.assertEqual(result.diff, ())

    @responses.activate
    def test_extras(self):
        """Update all extras"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file, '--all', '--no-base'))

        self.check_no_error(result)
        self.assertEqual(
            result.stdout,
            (
                'Base Dependencies:\n'
                'Optional Dependencies:\n'
                '  opt1:\n'
                '    <-- prefixed ~= 0.3.2\n'
                '    --> prefixed ~= 0.6.0\n'
                '  opt2:\n'
                '    <-- sphinx ~= 5.3\n'
                '    --> sphinx ~= 6.1\n'
                '  opt3:\n'
                '    <-- blessed < 1.16, >= 1.0\n'
                '    --> blessed < 1.19.1, >= 1.0\n'
                '  opt4:\n'
                '    <-- six == 1.7.1\n'
                '    --> six == 1.16.0\n'
            )
        )
        self.assertEqual(result.diff, DIFF_EXTRAS)

    @responses.activate
    def test_extras_specific(self):
        """Update specific extra"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file,  'opt2', 'opt5'))

        self.check_no_error(result)
        self.assertEqual(
            result.stdout,
            (
                'Base Dependencies:\n'
                'Optional Dependencies:\n'
                '  opt2:\n'
                '    <-- sphinx ~= 5.3\n'
                '    --> sphinx ~= 6.1\n'
            )
        )
        self.assertEqual(
            result.diff, (
                "-     'sphinx ~= 5.3',",
                '+     "sphinx ~= 6.1",',
            )
        )

    @responses.activate
    def test_regex_exclude(self):
        """Exclude dependencies with regex"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file,  '-e', 'req.*'))

        self.check_no_error(result)
        self.assertEqual(
            result.stdout,
            (
                'Base Dependencies:\n'
                '    <-- packaging == 1.0.0\n'
                '    --> packaging == 23.1\n'
                'Optional Dependencies:\n'
            )
        )
        self.assertEqual(
            result.diff, (
                "-     'packaging == 1.0.0',",
                '+     "packaging == 23.1",',
            )
        )

    @responses.activate
    def test_regex_include(self):
        """Include dependencies with regex"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file,  '-i', 'pack.*'))

        self.check_no_error(result)
        self.assertEqual(
            result.stdout,
            (
                'Base Dependencies:\n'
                '    <-- packaging == 1.0.0\n'
                '    --> packaging == 23.1\n'
                'Optional Dependencies:\n'
            )
        )
        self.assertEqual(
            result.diff, (
                "-     'packaging == 1.0.0',",
                '+     "packaging == 23.1",',
            )
        )


class TestMainErrors(MockedCLI):
    """
    Tests for CLI Errors
    """

    @responses.activate
    def test_extras_unknown(self):
        """Unknown extras provided"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f',  result.file, 'unknown'))

        self.assertEqual(result.exit_code, 9)
        self.assertEqual(len(result.logs), 1)
        self.assertEqual(
            result.logs[0], 'ERROR:bumpdeps:Unknown section for optional dependencies: unknown'
        )
        self.assertEqual(result.stderr, '')
        self.assertEqual(result.stdout, 'No updates required\n')
        self.assertEqual(result.diff, ())

    def test_file_not_found(self):
        """Bad file name results in error"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', 'NOT_A_REAL_FILE.toml'))

        self.assertEqual(result.exit_code, 2)
        self.assertEqual(len(result.logs), 0)
        self.assertRegex(result.stderr, 'File NOT_A_REAL_FILE.toml does not exist')
        self.assertEqual(result.stdout, '')
        self.assertEqual(result.diff, ())

    @responses.activate
    def test_ignore_until_format_invalid(self):
        """ignore-until format is not valid"""

        with self.write_and_diff(
            '[project]\ndependencies = [\n"six == 1.7.1",  # bumpdeps: ignore-until=tomorrow\n]\n'
        ) as result:
            bumpdeps.main(('-f',  result.file,))

        self.assertEqual(result.exit_code, 9)
        self.assertEqual(len(result.logs), 1)
        self.assertEqual(
            result.logs[0], 'ERROR:bumpdeps:Invalid format for ignore-until: ignore-until=tomorrow'
        )
        self.assertEqual(result.stderr, '')
        self.assertEqual(result.stdout, 'No updates required\n')
        self.assertEqual(result.diff, ())

    @responses.activate
    def test_ignore_until_date_invalid(self):
        """Data is not a valid date"""

        with self.write_and_diff(
            '[project]\ndependencies = [\n"six == 1.7.1",  # bumpdeps: ignore-until=2022-02-30\n]\n'
        ) as result:
            bumpdeps.main(('-f',  result.file))

        self.assertEqual(result.exit_code, 9)
        self.assertEqual(len(result.logs), 1)
        self.assertEqual(
            result.logs[0], 'ERROR:bumpdeps:Invalid date provided for ignore-until: 2022-02-30'
        )
        self.assertEqual(result.stderr, '')
        self.assertEqual(result.stdout, 'No updates required\n')
        self.assertEqual(result.diff, ())

    def test_regex_exclude_invalid(self):
        """Regex for exclude in invalid"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file, '-e', '('))

        self.assertEqual(result.exit_code, 2)
        self.assertEqual(len(result.logs), 0)
        self.assertRegex(result.stderr,  r"Invalid regex '\(' provided for --exclude")
        self.assertEqual(result.stdout, '')
        self.assertEqual(result.diff, ())

    def test_regex_include_invalid(self):
        """Regex for include in invalid"""

        with self.write_and_diff(EXAMPLE) as result:
            bumpdeps.main(('-f', result.file, '-i', '('))

        self.assertEqual(result.exit_code, 2)
        self.assertEqual(len(result.logs), 0)
        self.assertRegex(result.stderr,  r"Invalid regex '\(' provided for --include")
        self.assertEqual(result.stdout, '')
        self.assertEqual(result.diff, ())

    def test_requirement_invalid(self):
        """Requirement does not have valid format"""

        with self.write_and_diff(
            '[project]\ndependencies = [\n"requests !! 1.2.3" \n]\n'
        ) as result:
            bumpdeps.main(('-f', result.file))

        self.check_result_error(result, 'Invalid requirement')

    @responses.activate
    def test_response_error(self):
        """Bad response from package index"""

        with self.write_and_diff(
            '[project]\ndependencies = [\n"no_such_package == 1.2.3" \n]\n'
        ) as result:
            bumpdeps.main(('-f', result.file))

        self.check_result_error(result, 'Unable to query package index for no_such_package')

    @responses.activate
    def test_response_json_invalid(self):
        """Response from package index is not valid JSON"""

        with self.write_and_diff(
            '[project]\ndependencies = [\n"invalid_json == 1.2.3" \n]\n'
        ) as result:
            bumpdeps.main(('-f', result.file))

        self.check_result_error(result, 'Invalid JSON returned from package index')

    @responses.activate
    def test_response_json_unexpected(self):
        """Response from package index does not have expected structure"""

        with self.write_and_diff(
            '[project]\ndependencies = [\n"unexpected_json == 1.2.3" \n]\n'
        ) as result:
            bumpdeps.main(('-f', result.file))

        self.check_result_error(result, 'Unexpected JSON structure returned from package index')

    def test_toml_invalid(self):
        """Invalid TOML in file"""

        with self.write_and_diff('{"json": true}') as result:
            bumpdeps.main(('-f', result.file))

        self.check_result_error(result, f'Error loading {result.file}')

    def test_toml_project_missing(self):
        """TOML file does not have required 'project' section"""

        with self.write_and_diff('[foobar]\nfoo = "bar"\n') as result:
            bumpdeps.main(('-f', result.file))

        self.check_result_error(result, 'No project section in file')



# -*- coding: utf-8 -*-
# Copyright 2023 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**BumpDeps class Unit Tests**
"""

import responses

import bumpdeps
from tests import DIFF_BASE, DIFF_EXTRAS, EXAMPLE, MockedResponse, write_and_diff


class TestBumpDeps(MockedResponse):
    """
    Tests for updating dependencies
    """

    @responses.activate
    def test_base(self):
        """Update only base dependencies"""

        with write_and_diff(EXAMPLE) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            updates = bumper.bump()

        self.assertEqual(
            updates['dependencies'],
            [('packaging == 1.0.0', 'packaging == 23.1'), ('requests <= 2.0', 'requests <= 2.28.2')]
        )
        self.assertEqual(updates['optional-dependencies'], {})
        self.assertEqual(result.diff, DIFF_BASE)

    @responses.activate
    def test_base_no_deps(self):
        """No dependencies listed"""

        with write_and_diff('[project]\n') as result:
            bumper = bumpdeps.BumpDeps(result.file)
            updates = bumper.bump()

        self.assertEqual(updates['dependencies'], [])
        self.assertEqual(updates['optional-dependencies'], {})
        self.assertEqual(result.diff, tuple())

    @responses.activate
    def test_dry_run(self):
        """Updates returns, but file is not updated"""

        with write_and_diff(EXAMPLE) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            updates = bumper.bump(dry_run=True)

        self.assertEqual(
            updates['dependencies'],
            [('packaging == 1.0.0', 'packaging == 23.1'), ('requests <= 2.0', 'requests <= 2.28.2')]
        )

        self.assertEqual(result.diff, tuple())

    @responses.activate
    def test_extras(self):
        """Update all extras"""

        with write_and_diff(EXAMPLE) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            updates = bumper.bump(base=False, extras=True)

        self.assertEqual(updates['dependencies'], [])
        self.assertEqual(
            updates['optional-dependencies'],
            {
                'opt1': [('prefixed ~= 0.3.2', 'prefixed ~= 0.6.0')],
                'opt2': [('sphinx ~= 5.3', 'sphinx ~= 6.1')],
                'opt3': [('blessed < 1.16, >= 1.0', 'blessed < 1.19.1, >= 1.0')],
                'opt4': [('six == 1.7.1', 'six == 1.16.0')],
                'opt5': [],
            }
        )
        self.assertEqual(result.diff, DIFF_EXTRAS)

    @responses.activate
    def test_extras_specific(self):
        """Update specific extra"""

        with write_and_diff(EXAMPLE) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            updates = bumper.bump(base=False, extras=('opt2',))

        self.assertEqual(updates['dependencies'], [])
        self.assertEqual(
            updates['optional-dependencies'],
            {
                'opt2': [('sphinx ~= 5.3', 'sphinx ~= 6.1')],
            }
        )

        self.assertEqual(
            result.diff, (
                "-     'sphinx ~= 5.3',",
                '+     "sphinx ~= 6.1",',
            )
        )

    @responses.activate
    def test_regex_exclude(self):
        """Exclude dependencies with regex"""

        with write_and_diff(EXAMPLE) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            updates = bumper.bump(exclude='req.*')

        self.assertEqual(
            updates['dependencies'],
            [('packaging == 1.0.0', 'packaging == 23.1')]
        )
        self.assertEqual(updates['optional-dependencies'], {})

        self.assertEqual(
            result.diff, (
                "-     'packaging == 1.0.0',",
                '+     "packaging == 23.1",',
            )
        )

    @responses.activate
    def test_regex_include(self):
        """Include dependencies with regex"""

        with write_and_diff(EXAMPLE) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            updates = bumper.bump(include='pack.*')

        self.assertEqual(
            updates['dependencies'],
            [('packaging == 1.0.0', 'packaging == 23.1')]
        )
        self.assertEqual(updates['optional-dependencies'], {})

        self.assertEqual(
            result.diff, (
                "-     'packaging == 1.0.0',",
                '+     "packaging == 23.1",',
            )
        )


class TestBumpDepsErrors(MockedResponse):
    """
    Tests for errors raise in BumpDeps class
    """

    @responses.activate
    def test_extras_unknown(self):
        """Unknown extras provided"""

        with write_and_diff(EXAMPLE) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertLogs('bumpdeps', level='ERROR') as logs:
                updates = bumper.bump(base=False, extras=('unknown',))

        self.assertEqual(
            logs.output,
            ['ERROR:bumpdeps:Unknown section for optional dependencies: unknown'],
        )

        self.assertEqual(updates['dependencies'], [])
        self.assertEqual(updates['optional-dependencies'], {})
        self.assertEqual(result.diff, tuple())

    def test_file_not_found(self):
        """Bad file name raises error"""

        filename = 'NOT_A_REAL_FILE.toml'
        bumper = bumpdeps.BumpDeps(filename)
        with self.assertRaisesRegex(bumpdeps.BumpDepsError, f'Error loading {filename}'):
            bumper.bump()

    @responses.activate
    def test_ignore_until_format_invalid(self):
        """ignore-until format is not valid"""

        with write_and_diff(
            '[project]\ndependencies = [\n"six == 1.7.1",  # bumpdeps: ignore-until=tomorrow\n]\n'
        ) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertLogs('bumpdeps', level='ERROR') as logs:
                updates = bumper.bump()

        self.assertEqual(
            logs.output,
            ['ERROR:bumpdeps:Invalid format for ignore-until: ignore-until=tomorrow'],
        )

        self.assertEqual(updates['dependencies'], [])
        self.assertEqual(updates['optional-dependencies'], {})
        self.assertEqual(result.diff, tuple())

    @responses.activate
    def test_ignore_until_date_invalid(self):
        """Data is not a valid date"""

        with write_and_diff(
            '[project]\ndependencies = [\n"six == 1.7.1",  # bumpdeps: ignore-until=2022-02-30\n]\n'
        ) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertLogs('bumpdeps', level='ERROR') as logs:
                updates = bumper.bump()

        self.assertEqual(
            logs.output,
            ['ERROR:bumpdeps:Invalid date provided for ignore-until: 2022-02-30'],
        )

        self.assertEqual(updates['dependencies'], [])
        self.assertEqual(updates['optional-dependencies'], {})
        self.assertEqual(result.diff, tuple())

    def test_requirement_invalid(self):
        """Requirement does not have valid format"""
        with write_and_diff('[project]\ndependencies = [\n"requests !! 1.2.3" \n]\n') as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertRaisesRegex(bumpdeps.BumpDepsError, 'Invalid requirement'):
                bumper.bump()

    @responses.activate
    def test_response_error(self):
        """Bad response from package index"""
        with write_and_diff(
            '[project]\ndependencies = [\n"no_such_package == 1.2.3" \n]\n'
        ) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertRaisesRegex(
                bumpdeps.BumpDepsError, 'Unable to query package index for no_such_package'
            ):
                bumper.bump()

    @responses.activate
    def test_response_json_invalid(self):
        """Response from package index is not valid JSON"""
        with write_and_diff(
            '[project]\ndependencies = [\n"invalid_json == 1.2.3" \n]\n'
        ) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertRaisesRegex(
                bumpdeps.BumpDepsError, 'Invalid JSON returned from package index'
            ):
                bumper.bump()

    @responses.activate
    def test_response_json_unexpected(self):
        """Response from package index does not have expected structure"""
        with write_and_diff(
            '[project]\ndependencies = [\n"unexpected_json == 1.2.3" \n]\n'
        ) as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertRaisesRegex(
                bumpdeps.BumpDepsError,
                "Unexpected JSON structure returned from package index: {'version': '1.2.3'}"
            ):
                bumper.bump()

    def test_toml_invalid(self):
        """Invalid TOML in file"""
        with write_and_diff('{"json": true}') as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertRaisesRegex(bumpdeps.BumpDepsError, f'Error loading {result.file}'):
                bumper.bump()

    def test_toml_project_missing(self):
        """TOML file does not have required 'project' section"""
        with write_and_diff('[foobar]\nfoo = "bar"\n') as result:
            bumper = bumpdeps.BumpDeps(result.file)
            with self.assertRaisesRegex(bumpdeps.BumpDepsError, 'No project section in file'):
                bumper.bump()



# -*- coding: utf-8 -*-
# Copyright 2022 - 2023 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Bump Dependencies**

Utility for bumping dependency versions specified in pyproject.toml files
Attempts to adhere to specifications outlined in PEP 440 and PEP 508

While this author does not generally think upper pinning is appropriate in many
cases, there are some where it may be unavoidable. This utility is intended to
help in those cases to avoid dependency stagnation.

https://peps.python.org/pep-0440/
https://peps.python.org/pep-0508/
"""

import argparse
from collections import abc
from datetime import date as Date
from json import JSONDecodeError
import logging
from pathlib import Path
import re
import sys

import requests
import tomlkit
from tomlkit.exceptions import TOMLKitError
from packaging.requirements import Requirement, InvalidRequirement
from packaging.specifiers import Specifier, SpecifierSet
from packaging.utils import canonicalize_name
from packaging.version import Version


__version__ = '0.2.1'
__all__ = 'BumpDeps', 'BumpDepsError', 'main'

DESCRIPTION = "Utility for bumping dependencies in pyproject.toml files"
EPILOG = '''\
Upper pinning of dependencies is generally not a good idea outside of deployment.
This utility is intended for cases where it can't be avoided.
'''

LOGGER = logging.getLogger('bumpdeps')
LOGGER.addHandler(logging.NullHandler())

COMMENT_RE = re.compile(r'[\s#]*bumpdeps:\s*(.+)\s*$')
IGNORE_UNTIL_RE = re.compile(r'ignore-until\s*=\s*(\d{4}-\d{2}-\d{2})\s*$')


class BumpDepsError(Exception):
    """General exception class for BumpDep errors"""


def _dump_requirement(req):
    """
    Converts a Requirement to a string
    This is only needed because the default __str__ methods don't include spaces

    This can be removed if this issue is resolved
    https://github.com/pypa/packaging/issues/654
    """

    # pylint: disable=protected-access
    return ' '.join(filter(bool, (
        req.name,
        f'[{",".join(req.extras)}]' if req.extras else None,
        ', '.join(sorted(f'{spec.operator} {spec.version}' for spec in req.specifier._specs)),
        f'@ {req.url}' if req.url else None,
        f'; {req.marker}' if req.marker else None,
    )))


class PyPI:
    """
    Class for accessing package index
    Defaults to PyPI, but works with other compatible indexes
    """

    def __init__(self, base_url='https://pypi.org') -> None:
        self.base_url = base_url

    def get_latest_package_version(self, package):
        """
        Query latest package version from package index
        """

        url = '/'.join((self.base_url, 'pypi', canonicalize_name(package), 'json'))

        LOGGER.debug('Querying latest version for %s from %s', package, url)
        response = requests.get(url, headers={'Accept': 'application/json'}, timeout=5)

        try:
            response.raise_for_status()  # Improve error handling later
        except requests.HTTPError as e:
            raise BumpDepsError(
                f'Unable to query package index for {package}: {e} {response.text}'
            ) from e

        try:
            return response.json()['info']['version']
        except JSONDecodeError as e:
            raise BumpDepsError(f'Invalid JSON returned from package index: {e}') from e
        except KeyError as e:
            raise BumpDepsError(
                f'Unexpected JSON structure returned from package index: {response.json()}'
            ) from e


class BumpDeps:
    """
    Main class for bumping dependencies

    Defaults to using PYPI and pyproject.toml, but customizable
    """

    def __init__(self, filename, pkg_index=None) -> None:

        self.filepath = Path(filename)
        self.pkg_index = PyPI(pkg_index) if pkg_index else PyPI()

    def bump(self, base=True, extras=False, include=None, exclude=None, dry_run=False):
        """
        Bump dependencies

        If base is True, base dependencies are updated
        If extras is True, all optional dependencies are updated
        If extras is an iterable, only those optional dependencies will be updated

        """

        updates = {'dependencies': [], 'optional-dependencies': {}}

        LOGGER.debug('Loading configuration from %s', self.filepath)
        try:
            with self.filepath.open('r', encoding='utf-8') as config_file:
                config = tomlkit.load(config_file)
        except (OSError, TOMLKitError) as e:
            raise BumpDepsError(f'Error loading {self.filepath}: {e}') from e

        if 'project' not in config:
            raise BumpDepsError(f'No project section in file {self.filepath}')

        # Update base dependencies
        if base is True and 'dependencies' in config['project']:
            LOGGER.debug('Updating base dependencies')
            updates['dependencies'] = self._update_deps(
                config['project']['dependencies'], include=include, exclude=exclude
            )

        # Update optional dependencies
        opt_deps = config['project'].get('optional-dependencies')

        if opt_deps and extras is True:
            for extra, section in config['project']['optional-dependencies'].items():
                LOGGER.debug('Updating optional dependencies for extra %s', extra)
                updates['optional-dependencies'][extra] = self._update_deps(
                    section, include=include, exclude=exclude)

        elif opt_deps and isinstance(extras, abc.Iterable):
            for extra in extras:
                if extra not in opt_deps:
                    LOGGER.error('Unknown section for optional dependencies: %s', extra)
                    continue

                LOGGER.debug('Updating optional dependencies for extra %s', extra)
                updates['optional-dependencies'][extra] = self._update_deps(
                    opt_deps[extra], include=include, exclude=exclude)

        # If dry run, return changes
        if dry_run:
            LOGGER.debug('Dry run. No changes persisted')
            return updates

        # Write toml to temp file
        if updates['dependencies'] or any(updates['optional-dependencies'].values()):

            temp_path = self.filepath.with_suffix('._bumpdeps')
            LOGGER.debug("Writing output to temp file '%s'", temp_path)
            with temp_path.open('w', encoding='utf-8') as temp_file:
                tomlkit.dump(config, temp_file)

            LOGGER.debug("Moving '%s' to '%s'", temp_path, self.filepath)
            temp_path.replace(self.filepath)
        else:
            LOGGER.debug('No changes to persisted')

        return updates

    def _ignore_for_comment(self, comment):

        match = COMMENT_RE.search(comment)

        if match is None:
            return False

        directives = match[1]
        if 'ignore-until' in directives:
            date_match = IGNORE_UNTIL_RE.search(directives)
            if date_match is None:
                LOGGER.error('Invalid format for ignore-until: %s', directives)
            else:
                try:
                    return Date.fromisoformat(date_match[1]) > Date.today()

                except ValueError:
                    LOGGER.error('Invalid date provided for ignore-until: %s', date_match[1])

        # Skip if ignore
        return 'ignore' in directives

    def _update_deps(self, section, include=None, exclude=None):
        """
        Update dependencies for an extra section or base
        """

        updates = []

        # A little bit of a hack in order to be able to read comments
        for idx, spec in enumerate(section._value):  # pylint: disable=protected-access
            if not spec.value:  # tomlkit._ArrayItemGroup
                continue

            # Parse requirement
            try:
                req = Requirement(spec.value)
            except InvalidRequirement as e:
                raise BumpDepsError(f"Invalid requirement '{spec.value}': {e}") from e

            if exclude and re.match(exclude, req.name):
                LOGGER.debug('Skipping package %s for exclude filter: %s', req.name, exclude)
                continue

            if include and not re.match(include, req.name):
                LOGGER.debug('Skipping package %s for include filter: %s', req.name, include)
                continue

            # Check comments
            if spec.comment and self._ignore_for_comment(str(spec.comment)):
                LOGGER.debug('Skipping package %s for comment: %s', req.name, spec.comment)
                continue

            # No need to update if no specifiers are defined
            if not req.specifier:
                LOGGER.debug('Skipping package %s since no specifiers defined', req.name)
                continue

            # Get latest version
            latest = self.pkg_index.get_latest_package_version(req.name)
            LOGGER.debug('Latest version for package %s is %s', req.name, latest)

            # No need to update if latest version isn't precluded
            if next(req.specifier.filter((latest,)), False):
                LOGGER.debug('Skipping package %s since latest version applies', req.name)
                continue

            # Capture requirement before updating
            old_req = _dump_requirement(req)
            LOGGER.debug("Updating specifiers for requirement '%s'", old_req)

            # Update specifiers
            req.specifier = self._update_specifiers(req.specifier, latest)
            new_req = _dump_requirement(req)

            # Update section
            LOGGER.debug("Re-adding updated requirement '%s'", new_req)
            section[idx] = new_req
            updates.append((old_req, new_req))

        return updates

    def _update_specifiers(self, specifier_set, latest):
        """
        Update a specifier set for a dependency
        """

        resultant_specifiers = []
        for specifier in specifier_set:
            # No maximum version specified, so keep as is
            if specifier.operator in ('!=', '>=', '>'):
                resultant_specifiers.append(specifier)

            # Specific or max version specified, replace with the latest
            elif specifier.operator in ('==', '===', '<', '<='):
                version = Version(specifier.version)
                new = Specifier(f'{specifier.operator}{latest}')
                resultant_specifiers.append(new)

            # Compatibility operator, try to match places
            elif specifier.operator == '~=':
                version = Version(specifier.version)
                latest_version = Version(latest)

            # If maximum version is specified, see how many places are specified
                version = Version(specifier.version)

                # Less than 3 places are occupied and latest has at least 3, use 2
                if len(version.release) < 3 <= len(latest_version.release):
                    new = Specifier(f'~={".".join(map(str, latest_version.release[:2]))}')
                else:
                    new = Specifier(f'~={latest}')

                resultant_specifiers.append(new)

            else:  # pragma: no cover
                raise BumpDepsError(
                    f"Unexpected operator '{specifier.operator}' in '{specifier}'"
                )

        return SpecifierSet(','.join(map(str, resultant_specifiers)))


def cli(args=None):
    """
    Parse CLI arguments
    """

    parser = argparse.ArgumentParser(description=DESCRIPTION, epilog=EPILOG)

    parser.add_argument('-a', '--all', action='store_true', default=False,
                        help='Update all base and extra dependencies')
    parser.add_argument('-b', '--base', action='store_true', default=False,
                        help='Update base dependencies, default when no extras are provided')
    parser.add_argument('-n', '--no-base', action='store_true', default=False,
                        help='Do not update base dependencies, for use with --all')
    parser.add_argument('-i', '--include', metavar='REGEX',
                        help='Only include dependency names matching regex')
    parser.add_argument('-e', '--exclude', metavar='REGEX',
                        help='Exclude dependency names matching regex')
    parser.add_argument('-f', '--file', metavar='FILE', type=Path, default=Path('pyproject.toml'),
                        help="Path to TOML file, default: 'pyproject.toml'")
    parser.add_argument('--dry-run', action='store_true', default=False,
                        help='No changes made, but report what would have changed')
    parser.add_argument('--pkg-index', metavar='URL', default=None,
                        help='Package index URL, Default: https://pypi.org')
    parser.add_argument('extras', nargs='*', metavar='EXTRAS',
                        help='Extra dependency sections to bump, bump base if none provided')
    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help='Enable debug logging')

    options = parser.parse_args(args)

    if not options.file.is_file():
        parser.error(f'File {options.file} does not exist')

    if options.include:
        try:
            re.compile(options.include)
        except re.error as e:
            parser.error(f"Invalid regex '{options.include}' provided for --include: {e}")

    if options.exclude:
        try:
            re.compile(options.exclude)
        except re.error as e:
            parser.error(f"Invalid regex '{options.exclude}' provided for --exclude: {e}")

    log_level = logging.DEBUG if options.debug else logging.INFO
    logging.basicConfig(level=log_level)

    return options


def main(args=None):
    """
    Main entry point for CLI
    """

    options = cli(args)
    bumper = BumpDeps(options.file, options.pkg_index)
    try:
        updates = bumper.bump(
            (options.base or options.all or not options.extras) and not options.no_base,
            options.extras or options.all,
            include=options.include, exclude=options.exclude, dry_run=options.dry_run
        )
    except BumpDepsError as e:
        LOGGER.error('%s', e)
        sys.exit(8)

    # Notify user of updates
    if updates['dependencies'] or any(updates['optional-dependencies'].values()):

        print('Base Dependencies:')
        for entry in updates['dependencies']:
            print(f'    <-- {entry[0]}')
            print(f'    --> {entry[1]}')

        print('Optional Dependencies:')
        for extra, reqs in updates['optional-dependencies'].items():
            if not reqs:
                continue
            print(f'  {extra}:')
            for entry in reqs:
                print(f'    <-- {entry[0]}')
                print(f'    --> {entry[1]}')
    else:
        print('No updates required')

    # Set return value if an error was logged
    if LOGGER._cache.get(logging.ERROR):  # pylint: disable=protected-access
        sys.exit(9)

    sys.exit(0)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2022 - 2023 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Nox configuration file for BumpDeps
See https://nox.thea.codes/en/stable/config.html
"""

from pathlib import Path

import nox
import tomli

BASE_PYTHON = '3.10'

with open('pyproject.toml', 'rb') as toml_file:
    CONFIG = tomli.load(toml_file)

DEPENDENCIES = CONFIG['project']['dependencies']
TEST_DEPENDENCIES = CONFIG['project']['optional-dependencies']['tests']
NOX_DEPENDENCIES = ("nox", "tomli")

# Nox options
nox.needs_version = ">=2022.8.7"
nox.options.stop_on_first_error = False
nox.options.error_on_missing_interpreters = False


@nox.session(python=BASE_PYTHON, tags=['lint'])
def pylint(session: nox.Session) -> None:
    """Run Pylint"""
    session.install('pylint', 'pyenchant', *NOX_DEPENDENCIES, *DEPENDENCIES, *TEST_DEPENDENCIES)
    session.run('pylint', 'bumpdeps', 'noxfile.py', 'tests')


@nox.session(python=BASE_PYTHON, tags=['lint'])
def flake8(session: nox.Session) -> None:
    """Run Flake8"""
    session.install('flake8')
    session.run('flake8', '--max-line-length', '100', '--benchmark')


@nox.session(python=BASE_PYTHON, tags=['lint'])
def readme(session: nox.Session) -> None:
    """Check readme"""

    Path('build').mkdir(exist_ok=True)
    session.install('rst2html', 'Pygments')
    session.run('rst2html.py', '-v', '--strict', 'README.rst', 'build/README.html', external=True)


@nox.session(python=['3.7', '3.8', '3.9', '3.10', '3.11'], tags=['test'])
def test(session: nox.Session) -> None:
    """Run unit tests"""
    session.install('.[tests]')
    session.run('python', '-m', 'unittest', 'discover', '-s', 'tests', *session.posargs)


@nox.session(python=BASE_PYTHON, tags=['test'])
def coverage(session: nox.Session) -> None:
    """Run unit tests"""
    session.install('.[tests]', 'coverage[toml]')
    session.run('coverage', 'erase')
    session.run('coverage', 'run', '-m', 'unittest', 'discover', '-s', 'tests', *session.posargs)
    session.run('coverage', 'report')
